from scrapeops_scrapy.signals import scrapeops_signals

class ScrapeOpsTrigger(object):

    def __init__(self):
        pass
    
    @staticmethod
    def reject_response(crawler=None, response=None, reason=None):   
        crawler.signals.send_catch_log(signal=scrapeops_signals.scrapeops_response_rejected, 
                spider=crawler.spider, 
                response=response,
                reason=reason,
                )

    @staticmethod
    def reject_item(crawler=None, response=None, item=None, reason=None):   
        crawler.signals.send_catch_log(signal=scrapeops_signals.scrapeops_item_rejected, 
                spider=crawler.spider, 
                response=response,
                item=item,
                reason=reason,
                )
    

    
 



   


    







         
    



from scrapeops_scrapy.core.api import SOPSRequest 
from scrapeops_scrapy.normalizer.domains import DomainNormalizer
from scrapeops_scrapy.utils import utils
import json
import logging
import re

class ErrorLogger(object):

    ERROR_LOGGER_ACTIVE = True

    def __init__(self, spider, crawler, spider_settings, server_hostname, server_ip, start_time, log_file):
        self.spider = spider
        self.crawler = crawler
        self.bot_name = crawler.settings.get('BOT_NAME', 'None')
        self.spider_settings = spider_settings
        self.server_hostname = server_hostname
        self.server_ip = server_ip
        self.start_time = start_time
        self.log_file = log_file
        self._error_history = []
        self.job_group_name = None
        self.job_group_id = None

    def update_error_logger(self, job_name, job_id):
        self.job_group_name = job_name
        self.job_group_id = job_id

    def log_error(self, reason=None, error=None, data=None, request_type=None):
        if ErrorLogger.ERROR_LOGGER_ACTIVE:
            self._error_history.append({
                'time': utils.current_time(),
                'reason': reason,
                'error': str(error),
                'data': data,
                'request_type': request_type,
            })
    

    def send_error_report(self, error_type=None, body=None, log_data=False):
        if ErrorLogger.ERROR_LOGGER_ACTIVE:
            try:
                data, status = SOPSRequest().error_report_request(error_type=error_type, body=body)
                if status.valid:
                    if log_data and self.log_file is not None and data.get('sdk_error_id') is not None:
                        with open(self.log_file, 'rb') as f:
                            post_body = {
                                            'sops_sdk': 'scrapy',
                                            'spider_name': self.spider.name,
                                            'job_group_id': self.job_group_id,
                                            'job_group_name': self.job_group_name,
                                            'sdk_error_id': data.get('sdk_error_id')
                                        }
                            _, status = SOPSRequest().error_report_request(error_type=error_type, body=post_body, files={'file': f})  
                            if status.valid is False:
                                self.log_error(reason='send_error_logs_failed', error=status.error)  

                if status.valid is False:
                        self.log_error(reason='send_error_report_failed', error=status.error) 
            except Exception:
                pass


    def sdk_error_close(self, reason=None, error=None, request_type=None, data=None):
        if ErrorLogger.ERROR_LOGGER_ACTIVE:
            self.log_error(reason=reason, error=error, data=data, request_type=request_type)
            error_data = {
                'final_reason': reason,
                'sops_sdk': 'scrapy',
                'spider_name': self.spider.name,
                'bot_name': self.bot_name, 
                'server_ip': self.server_ip,
                'server_hostname': self.server_hostname,
                'job_group_id': self.job_group_id,
                'job_group_name': self.job_group_name,
                'job_args': utils.get_args(),
                'job_start_time': self.start_time,
                'sops_scrapeops_version': utils.get_scrapeops_version(),
                'sops_scrapy_version': utils.get_scrapy_version(),
                'sops_python_version': utils.get_python_version(),
                'sops_system_version': utils.get_system_version(),
                'sops_middleware_enabled': utils.scrapeops_middleware_installed(self.spider_settings),
                'error_history': self._error_history,
            }
            
            self.send_error_report(error_type='sdk_close', body=error_data, log_data=True)
    

         

class TailLogHandler(logging.Handler):

    retryErrors = [
        "Couldn't bind",
        "Hostname couldn't be looked up'"
        "No route to host",
        "Connection was refused by other side",
        "TCP connection timed out",
        "File used for UNIX socket is no good",
        "Service name given as port is unknown",
        "User aborted connection",
        "User timeout caused connection failure",
        "An SSL error occurred",
        "Could not verify something that was supposed to be signed.",
        "The peer rejected our verify error.",
        "We did not find a certificate where we expected to find one.",
        "Bad Request",
        "Unauthorized",
        "Payment Required",
        "Forbidden",
        "Not Found",
        "Method Not Allowed",
        "Request Time-out",
        "Internal Server Error",
        "Bad Gateway",
        "Service Unavailable",
        "HTTP Version not supported",
        "Gateway Time-out",
        "Unknown Status",
    ]

    def __init__(self, log_dict, log_dict_cumulative):
        logging.Handler.__init__(self)
        self.log_dict = log_dict
        self.log_dict_cumulative = log_dict_cumulative


    def flush(self):
        self.log_dict.clear()
        

    def emit(self, record):

        try:
           
            if(record.levelname == "ERROR" or record.levelname == "WARNING" or record.levelname == "CRITICAL"):

                errorMessage = record.message
                fileAndLine = record.pathname + ', line: ' + str(record.lineno)
                dateTime = record.asctime
                type = record.levelname
                engine = record.name


                #covering warnings/probableCause/traceback missing
                traceback = 'No traceback available'
                probableCause = ''

                if record.exc_text is not None:
                    traceback = record.exc_text
                    splitTraceback = traceback.split('\n')
                    probableCause = splitTraceback[len(splitTraceback) - 1]


                #covering retrys
                if("Gave up retrying <" in record.message):

                    for retryError in self.retryErrors:
                        if(retryError in record.message):
                            method = record.message.split('<')[1].split(' ')[0]
                            errorMessage = "Error: Gave up retrying " + method + " request - " + retryError
                            fileAndLine = ''
                            probableCause = retryError
                            break
                
                # Deprecation Warnings
                if "ScrapyDeprecationWarning:" in record.message and record.message[0] == "/":
                    splitString = record.message.split("ScrapyDeprecationWarning:")
                    errorMessage = "ScrapyDeprecationWarning: " + splitString[1]
                    probableCause = splitString[0]


                # "Some Other Error Occurred"
                if "Some other error occurred: " in record.message: 
                    splitError = record.message.split(' /')
                    cleanError = splitError[0].split(">: ")[1]
                    errorMessage = "Some other error occurred: " + cleanError
                    probableCause = cleanError
                    traceback = record.message


                # Convert Urls To Domains in Error Messages
                urls = re.findall(r'(https?://[^\s]+)', errorMessage)
                for url in urls:
                    domain = DomainNormalizer.get_domain(url)
                    errorMessage = errorMessage.replace(url, domain)


                if errorMessage in self.log_dict:
                    self.log_dict[errorMessage]['count'] = self.log_dict[errorMessage]['count'] + 1
                else:
                    self.log_dict[errorMessage] = {
                        'type': type,
                        'engine': engine,
                        'name': errorMessage,
                        'count': 1, 
                        'traceback': traceback, 
                        'message' : probableCause, 
                        'filepath': fileAndLine, 
                        'dateTime': dateTime
                        }

                if(SOPSRequest.HIGH_FREQ_ACC == True):

                    if(errorMessage in self.log_dict_cumulative):
                        self.log_dict_cumulative[errorMessage]['count'] = self.log_dict_cumulative[errorMessage]['count'] + 1
                    else:

                        self.log_dict_cumulative[errorMessage] =  {
                            'type': type,
                            'engine': engine,
                            'name': errorMessage,
                            'count': 1, 
                            'traceback': traceback, 
                            'message' : probableCause, 
                            'filepath': fileAndLine, 
                            'dateTime': dateTime
                        }
                        
        except Exception as e:
            logging.info('Error: Error in error logger')
            logging.info(e, exc_info=True)

class TailLogger(object):

    def __init__(self):
        self._log_dict = {}
        self._log_dict_cumulative = {}
        self._log_handler = TailLogHandler(self._log_dict, self._log_dict_cumulative)

    def contents(self, type = "diff"):

        if(type == "cumulative"):
            jsonLogsCumulative = json.dumps(self._log_dict_cumulative, indent= 2)
            return jsonLogsCumulative

        else:
            jsonLogs = json.dumps(self._log_dict, indent= 2)
            self._log_handler.flush()
            return jsonLogs

    @property
    def log_handler(self):
        return self._log_handler





class BaseStatsModel(object):

    def __init__(self):
        pass

    def get_value(self, stats, key, default=None):
        return stats.get(key, default)

    def set_value(self, stats, key, value):
        stats[key] = value

    def inc_value(self, stats, key, count=1, start=0, spider=None):
        d = stats
        d[key] = d.setdefault(key, start) + count

    def max_value(self, stats, key, value, spider=None):
        stats[key] = max(stats.setdefault(key, value), value)

    def min_value(self, stats, key, value, spider=None):
        stats[key] = min(stats.setdefault(key, value), value)

    def print_stats(self, statsType, stats):
        print(f'#### SCRAPEOPS {statsType.upper()} STATS ####')
        print('{')
        for key, value in stats.items():
            if key[0] != '_':
                print(f"    '{key}': {value},")
        print('}')


class PeriodicStatsModel(BaseStatsModel):
    
    def __init__(self):
        self._periodic_stats = {}
        self._periodic_errors = 0
        self._periodic_warnings = 0
        self._periodic_criticals = 0

    def get_periodic_stats(self):
        return self._periodic_stats

    def reset_periodic_stats(self):
        self._periodic_stats = {}

    def display_periodic_stats(self):
        stats = self.get_periodic_stats()
        self.print_stats('periodic', stats)


class OverallStatsModel(BaseStatsModel):
    
    def __init__(self):
        self._overall_stats = {}
        self._overall_errors = 0
        self._overall_warnings = 0
        self._overall_criticals = 0

    def get_overall_stats(self):
        return self._overall_stats

    def display_overall_stats(self):
        stats = self.get_overall_stats()
        self.print_stats('overall', stats)




class ExceptionNormalizer(object):

    def __init__(self):
        pass
    
    @staticmethod
    def normalise_exception(exception_class):
        
        if 'ResponseNeverReceived' in exception_class:
            return 'ResponseNeverReceived'

        if 'Timeout' in exception_class:
            return 'Timeout'
        
        if 'TimedOut' in exception_class:
            return 'Timeout'

        if 'PotentialDataLoss' in exception_class:
            return 'PotentialDataLoss'

        if 'ConnectionLost' in exception_class:
            return 'ConnectionLost'

        return exception_class

    

    
 



   


    







         
    



## scrapy
from scrapy.utils.request import request_httprepr

## scrapeops
from scrapeops_scrapy.stats.model import OverallStatsModel, PeriodicStatsModel 
from scrapeops_scrapy.utils import utils
from scrapeops_scrapy.normalizer.exceptions import ExceptionNormalizer
from scrapeops_scrapy.utils.utils import get_header_size, get_status_size

import copy


class StatsLogger(OverallStatsModel, PeriodicStatsModel):

    def __init__(self):
        OverallStatsModel.__init__(self)
        PeriodicStatsModel.__init__(self)


    def display_stats(self):
        self.display_periodic_stats()
        self.display_overall_stats()


    def check_periodic_stats(self):
        if self._periodic_stats == {}:
            self.set_value(self._periodic_stats, 'job_id', self.job_id)


    def spider_open_stats(self):
        self.set_value(self._overall_stats, 'job_id', self.job_id)
        self.set_value(self._overall_stats, 'job_name', self.job_group_name)
        self.set_value(self._overall_stats, 'job_start_time', self.start_time)
        self.set_value(self._overall_stats, 'job_finish_time', 0)
        self.set_value(self._overall_stats, 'job_run_time', 0)
        self.set_value(self._overall_stats, 'status', 'Live')
        self.set_value(self._overall_stats, 'middleware_enabled', self._scrapeops_middleware)
        


    def spider_close_stats(self, reason=None, crawler=None):
        finish_time = utils.current_time()
        self.aggregate_stats(crawler) 
        self.set_value(self._overall_stats, 'job_finish_time', finish_time)
        self.set_value(self._overall_stats, 'job_run_time', finish_time - self.start_time)
        self.set_value(self._overall_stats, 'status', 'Finished')
        self.set_value(self._overall_stats, 'reason', reason)
        self.set_value(self._overall_stats, 'period_frequency', self._period_frequency)
        

    def generate_request_stats(self, request_response_object, request=None):
        proxy_name = request_response_object.get_proxy_name()
        proxy_setup = request_response_object.get_proxy_setup()
        domain_name = request_response_object.get_domain()
        page_type = request_response_object.get_page_type()
        custom_tag = request_response_object.get_custom_tag()
        reqlen = len(request_httprepr(request))
        
        ## periodic stats
        self.check_periodic_stats()
        self.inc_value(self._periodic_stats, f'requests|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{custom_tag}|count')
        self.inc_value(self._periodic_stats, f'requests|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{custom_tag}|bytes', count=reqlen)

        ## overall stats
        self.inc_value(self._overall_stats, f'requests|{request.method}|count')
        self.inc_value(self._overall_stats, f'requests|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{custom_tag}|count')
        self.inc_value(self._overall_stats, f'requests|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{custom_tag}|bytes', count=reqlen)


    def generate_response_stats(self, request_response_object, request=None, response=None):
        proxy_name = request_response_object.get_proxy_name()
        proxy_setup = request_response_object.get_proxy_setup()
        domain_name = request_response_object.get_domain()
        page_type = request_response_object.get_page_type()
        validation = request_response_object.get_validation_test()
        geo = request_response_object.get_geo()
        custom_tag = request_response_object.get_custom_tag()
        custom_signal = 'none'
        reslen = len(response.body) + get_header_size(response.headers) + get_status_size(response.status) + 4
        total_latency = request.meta.get('download_latency', 0)

        ## periodic stats
        self.check_periodic_stats()
        self.inc_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|count')
        self.inc_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|bytes', count=reslen)
        self.inc_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|retries', count=request.meta.get('retry_times', 0))
        self.inc_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|total_latency', count=total_latency)
        self.min_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|min_latency', total_latency)
        self.max_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|max_latency', total_latency)

        ## overall stats
        self.inc_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|count')
        self.inc_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|bytes', count=reslen)
        self.inc_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|retries', count=request.meta.get('retry_times', 0))
        self.inc_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|total_latency', count=total_latency)
        self.min_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|min_latency', total_latency)
        self.max_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{response.status}|{validation}|{geo}|{custom_tag}|{custom_signal}|max_latency', total_latency)


    def generate_item_stats(self, request_response_object, signal=None, response=None):
        if response is not None:
             request = response.request
             request_method = request.method
             status = response.status
        else: 
            request_method = status = 'unknown'
        proxy_name = request_response_object.get_proxy_name()
        proxy_setup = request_response_object.get_proxy_setup()
        domain_name = request_response_object.get_domain()
        page_type = request_response_object.get_page_type()
        validation = request_response_object.get_validation_test()
        geo = request_response_object.get_geo()
        custom_tag = request_response_object.get_custom_tag()
        custom_signal = 'none'
        self.check_periodic_stats()

        if signal == 'item_scraped':
            self.inc_value(self._periodic_stats, f'responses|{request_method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{status}|{validation}|{geo}|{custom_tag}|{custom_signal}|items')
            self.inc_value(self._overall_stats, f'responses|{request_method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{status}|{validation}|{geo}|{custom_tag}|{custom_signal}|items')
        
        elif signal == 'item_dropped':
            self.inc_value(self._periodic_stats, f'responses|{request_method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{status}|{validation}|{geo}|{custom_tag}|{custom_signal}|items_dropped')
            self.inc_value(self._overall_stats, f'responses|{request_method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{status}|{validation}|{geo}|{custom_tag}|{custom_signal}|items_dropped')
        
        elif signal == 'item_error':
            self.inc_value(self._periodic_stats, f'responses|{request_method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{status}|{validation}|{geo}|{custom_tag}|{custom_signal}|item_errors')
            self.inc_value(self._overall_stats, f'responses|{request_method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{status}|{validation}|{geo}|{custom_tag}|{custom_signal}|item_errors')

    

    def generate_exception_stats(self, request_response_object, request=None, exception_class=None):
        proxy_name = request_response_object.get_proxy_name()
        proxy_setup = request_response_object.get_proxy_setup()
        domain_name = request_response_object.get_domain()
        page_type = request_response_object.get_page_type()
        validation = request_response_object.get_validation_test()
        geo = request_response_object.get_geo()
        custom_tag = request_response_object.get_custom_tag()
        custom_signal = 'none'
        exception_type = ExceptionNormalizer.normalise_exception(exception_class)
        download_latency = request.meta.get('download_latency', 0)
        if download_latency is None:
            start_time = request.meta.get('sops_time', 0) 
            if start_time != 0: download_latency = utils.current_time() - start_time
            else: download_latency = 0 

        self.check_periodic_stats()
        self.inc_value(self._periodic_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{exception_type}|{validation}|{geo}|{custom_tag}|{custom_signal}|count')
        self.inc_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{exception_type}|{validation}|{geo}|{custom_tag}|{custom_signal}|count')
        self.inc_value(self._overall_stats, f'responses|{request.method}|{proxy_name}|{proxy_setup}|{domain_name}|{page_type}|{exception_type}|{validation}|{geo}|{custom_tag}|{custom_signal}|total_latency', count=download_latency)
    
    def aggregate_stats(self, crawler=None, middleware=False):
        self.avg_latency()
        self.log_levels(crawler)
        if middleware is False:
            self.get_exception_stats(crawler)


    def avg_latency(self):
        for stat_type in [self._periodic_stats, self._overall_stats]:
            stats_copy = copy.deepcopy(stat_type)
            for key, value in stats_copy.items():
                if 'responses' in key and 'total_latency' in key:
                    count_key = key.replace('total_latency', 'count')
                    avg_latency = value / stats_copy.get(count_key)
                    self.set_value(stat_type, key.replace('total_latency', 'avg_latency'), avg_latency)


    def log_levels(self, crawler):
        scrapy_stats = crawler.stats.get_stats()
        for log_level in ['WARNING', 'ERROR', 'CRITICAL']:
            log_key = 'log_count/' + log_level
            log_value = scrapy_stats.get(log_key, 0)
            previous_value = self._overall_stats.get(log_key, 0)
            self.set_value(self._periodic_stats, log_key, log_value - previous_value)
            self.set_value(self._overall_stats, log_key, log_value)

    
    def exception_type_check(self, key):
        if isinstance(key, str):
            return key.startswith('downloader/exception_type_count/')
        return False
    
    def get_exception_stats(self, crawler):
        scrapy_stats = crawler.stats.get_stats()
        if scrapy_stats.get('downloader/exception_count') is not None:
            exception_values = [ {k:v} for k,v in scrapy_stats.items() if self.exception_type_check(k)]
            for exception in exception_values:
                for key, value in exception.items():
                    key_type = key.replace('downloader/exception_type_count/', '')
                    try:
                        exception_type = key_type.split('.')[-1]
                    except Exception:
                        exception_type = key_type
                    self.set_value(self._overall_stats, f'responses|unknown|unknown|unknown|unknown|unknown|{exception_type}|unknown|unknown|unknown|unknown|count', value)

    

            

    
                


    








    
        





import functools

from scrapeops_scrapy.exceptions import ScrapeOpsAPIResponseError

def exception_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ScrapeOpsAPIResponseError as e:
            pass
        except Exception as e:
            pass
    return wrapper


